// file  Tutorial1.cpp

// This is the main C++ source file for my Direct3D Tutorial #1 which
// demonstrates the importance of the "painter's algorithm" when using
// alpha blending to accomplish transparent textures on a Direct3D model.
// The particular model I use Direct3D to draw is the Munsell color solid.
// The purpose of this Direct3D tutorial is further explained on my
// web page at:
//    http://www.computersciencelab.com/Direct3DTut1.htm

// Because this code employs Direct3D 8, you must link with D3D8.LIB,
// D3DX8.LIB, and DXERR8.LIB.  Make sure that your compiler searches
// for .H and .LIB files in the DirectX 8 SDK before it searches any
// other location (because the compiler probably shipped with older
// versions of the DirectX files).

// You can check out my other software at:
//   www.computersciencelab.com

//                                           John Kopplin  1/10/03

                                          
//////////////////////////////////////////////////////////////////////////////
// Includes
//
//////////////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#define STRICT

#ifdef UNICODE
 #define _UNICODE
#endif

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
//#include <stdarg.h>

#include <d3d8.h>
#include <d3dx8.h>
#include <dxerr8.h>
#include <d3dxerr.h>

#include "d3dfont.h"             // needed for Microsoft's CD3DFont class

#include "cMunsell.h"            // needed for my cMunsell class


//////////////////////////////////////////////////////////////////////////////
// Defines and Typedefs
//
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Globals
//
//////////////////////////////////////////////////////////////////////////////

TCHAR       g_szAppName[] = TEXT( "Direct3DTut1" );

IDirect3D8        * g_pD3D       = NULL;
IDirect3DDevice8  * g_pd3dDevice = NULL;

cMunsell    * g_pMunsell = NULL;      // class that models Munsell color solid
CD3DFont    * g_pFont    = NULL;      // font for writing on a Direct3D window


//////////////////////////////////////////////////////////////////////////////
// Function Prototypes
//
//////////////////////////////////////////////////////////////////////////////

LRESULT CALLBACK WndProc( HWND hWnd, UINT msg,
                          WPARAM wParam, LPARAM lParam );

HRESULT          InitDirect3D( HWND hWndFocus );
void             ShutdownDirect3D();
HRESULT          RestoreDeviceObjects();
HRESULT          Render();
HRESULT          Resize3DEnvironment();

int CDECL        MessageBoxFmt( HWND hWnd, const TCHAR * pszTitle,
                                UINT uiStyle, const TCHAR * pszFormat, ... );


//////////////////////////////////////////////////////////////////////////////
// _tWinMain()
//
// All WIN32 executables begin execution in this WinMain() function.
//////////////////////////////////////////////////////////////////////////////

int WINAPI _tWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
                      LPTSTR pszCmdLine, int nCmdShow )
{
    //  Define the window class used for the main window.

    WNDCLASS    wc;

    ZeroMemory( &wc, sizeof( WNDCLASS ) );

    wc.style         = CS_HREDRAW | CS_VREDRAW;
                                            // invalidate the entire client
                                            //   area whenever its vertical
                                            //   or horizontal size changes
    wc.lpfnWndProc   = WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon( NULL, IDI_APPLICATION );
    wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
    wc.hbrBackground = NULL;
            // When the .hbrBackground member is NULL, an application
            // must paint its own background whenever it is requested
            // to paint in its client area.  To determine whether the
            // background must be painted, an application can either
            // process the WM_ERASEBKGND message or test the fErase
            // member of the PAINTSTRUCT structure filled by the
            // BeginPaint() function.
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_szAppName;

    //  Register this window class with the Windows OS.

    if ( ! RegisterClass( &wc ) )           // Petzold 5th Ed. p 427 shows 
    {                                       //  RegisterClass() to be superior
                                            //  to RegisterClassEx()

        //  One way to cause this failure is to compile this program
        //  with UNICODE defined [hence RegisterClass() was mapped to
        //  RegisterClassW()] but then run it on Windows 95/98.
        //  In this situation we must be careful to only employ
        //  ASCII versions of functions while displaying the error message:

        MessageBoxA( NULL,
                     "FATAL ERROR: RegisterClass() failed in _tWinMain()",
                     g_szAppName,
                     MB_ICONERROR );
        PostQuitMessage( 0 );
        return 0;
    }

    // Create a window.

    HWND  hWnd = CreateWindowEx(
                     WS_EX_OVERLAPPEDWINDOW,    // extended style
                     g_szAppName,               // window class
                     g_szAppName,               // window caption
                     WS_OVERLAPPEDWINDOW,       // window style
                     CW_USEDEFAULT,             // x
                     CW_USEDEFAULT,             // y
                     CW_USEDEFAULT,             // cx
                     CW_USEDEFAULT,             // cy
                     HWND_DESKTOP,              // parent HWND
                     NULL,                      // use the class menu
                     hInstance,                 // instance
                     NULL );                    // creation parameters

    // This call to CreateWindowEx() generates a WM_CREATE message to our
    // MainWndProc().

    if ( ! hWnd )
    {
        MessageBox( NULL,
                    TEXT( "ERROR: CreateWindow() failed in _tWinMain()" ),
                    g_szAppName,
                    MB_ICONERROR );
        PostQuitMessage( 0 );
        return 0;
    }

    // We have to wait until after the window has been created to
    // initialize Direct3D because the IDirect3D8::CreateDevice()
    // function will return an error if it is called during the handling
    // of the WM_CREATE message [the error is HRESULT = 0x88760827 =
    // MAKE_D3DHRESULT(2087)].

    InitDirect3D( hWnd );

    // Construct the font that we use to write upon the Direct3D window.

    while ( ( g_pFont = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD ) ) == NULL)
    {
        int   nCode = MessageBox( NULL,
                                  TEXT( "ERROR: insufficient memory.\n" )
                                  TEXT( "Close other running applications\n" )
                                  TEXT( "and then hit 'Retry'" ),
                                  g_szAppName,
                                  MB_RETRYCANCEL | MB_ICONERROR );
        if ( IDCANCEL == nCode )
        {
            PostQuitMessage( 0 );
            return 0;
        }
    }

    // Initialize the font's internal textures.

    HRESULT    hr;

    hr = g_pFont->InitDeviceObjects( g_pd3dDevice );

    if ( FAILED( hr ) )
        goto Exit;

    // Construct an instance of the cMunsell class which models the
    // Munsell color solid.

    while ( ( g_pMunsell = new cMunsell() ) == NULL )
    {
        int   nCode = MessageBox( NULL,
                                  TEXT( "ERROR: insufficient memory.\n" )
                                  TEXT( "Close other running applications\n" )
                                  TEXT( "and then hit 'Retry'" ),
                                  g_szAppName,
                                  MB_RETRYCANCEL | MB_ICONERROR );
        if ( IDCANCEL == nCode )
        {
            PostQuitMessage( 0 );
            return 0;
        }
    }

    hr = RestoreDeviceObjects();         // includes a call to
                                         //   cMunsell::CreateGeometry()

    if ( FAILED( hr ) )
        goto Exit;

    // Unless we request WS_VISIBLE, the CreateWindow() function creates
    // a hidden window.  The following call to ShowWindow() adds the
    // WS_VISIBLE style to the window class (if so commanded by the passed
    // parameter nCmdShow).

    ShowWindow( hWnd, nCmdShow );         // causes first WM_SIZE msg

    // WM_PAINT messages have the lowest priority of all windows messages.
    // They are sent only when the update region is not empty and NO
    // other messages are in the queue.  Waiting for a WM_PAINT message
    // is called "asynchronous painting".  But at this point in this
    // program we want to accomplish "synchronous painting", i.e. force
    // an immediate WM_PAINT message regardless of the presence of any
    // other messages in the message queue.  This is accomplished by the
    // following call to UpdateWindow().

    UpdateWindow( hWnd );                 // causes first WM_PAINT msg

    // We now enter the all-important message loop, which continually checks
    // if the Windows OS has a message available for any of the windows
    // owned by this application.  Note that only one window from among
    // all the applications currently running is the "active window"
    // (indicated by a highlighted caption bar).  Keystrokes are
    // dispatched to the window with "keyboard focus", which is either
    // the active window or a descendant of the active window.

    MSG   msg;

    while ( GetMessage( &msg, NULL, 0, 0 ) )
    {
        // Translate raw keystrokes to higher level character messages.

        TranslateMessage( &msg );

        // Send the message to the appropriate window procedure (each
        // window belongs to a window class that specifies the window
        // procedure for all windows built from that class).

        DispatchMessage( &msg );
    }

Exit:
    hr = g_pFont->InvalidateDeviceObjects();

    hr = g_pFont->DeleteDeviceObjects();

    hr = g_pMunsell->ReleaseGeometry(); // releases vertex and index buffers

    ShutdownDirect3D();

    SAFE_DELETE( g_pFont );
    SAFE_DELETE( g_pMunsell );

    UnregisterClass( g_szAppName, hInstance );

    return 0;

} // _tWinMain()


//////////////////////////////////////////////////////////////////////////////
// WndProc()
//
//////////////////////////////////////////////////////////////////////////////

LRESULT CALLBACK WndProc( HWND hWnd, UINT msg,
                          WPARAM wParam, LPARAM lParam )
{
    // Window procedures exist to receive the windows messages sent by
    // the message loop in WinMain().  All windows constructed from the
    // same window class share the same window procedure.

    switch ( msg )
    {
        case WM_PAINT:
        {
            PAINTSTRUCT    ps;

            HDC      hDC = BeginPaint( hWnd, &ps );

            if ( g_pd3dDevice )
                g_pd3dDevice->Present( NULL, NULL, NULL, NULL );

            EndPaint( hWnd, &ps );

            return 0;
        }
        case WM_SIZE:
        {
            UINT    uiSizeType = wParam;            // SIZE_RESTORED, etc.
            UINT    uiClientWidth  = LOWORD( lParam );
            UINT    uiClientHeight = HIWORD( lParam );

            // A new window size will require a new backbuffer size
            // so the 3D structures must be changed accordingly.

            if ( uiSizeType == SIZE_MINIMIZED )
                return 0;

            // Resize the 3D environment

            HRESULT    hr;

            hr = Resize3DEnvironment();

            if ( FAILED( hr ) )
                MessageBox( NULL,
                            TEXT( "Resize3DEnvironment() failed" ),
                            g_szAppName,
                            MB_ICONERROR );
            hr = Render();

            return 0 ;
        }
        case WM_KEYDOWN:
        {
            switch( wParam )
            {
            case VK_ESCAPE:
                PostMessage( hWnd, WM_CLOSE, 0, 0 );
                break;
            }

            return 0;
        }
        case WM_CHAR:
        {
            switch ( wParam )
            {
            case 'b':
              {
                g_pMunsell->ToggleBackfaceCulling();
                break;
              }
            case 't':
              {
                g_pMunsell->ToggleTransparency();
                break;
              }
            case 'p':
              {
                g_pMunsell->TogglePaintersAlgorithm();
                break;
              }
            case 'r':
              {
                float    fRotation = g_pMunsell->GetRotation();
                fRotation += 0.1f;             // in rads
                if ( fRotation > TWOPI )
                    fRotation = (float)(fRotation - TWOPI);
                g_pMunsell->SetRotation( fRotation );
                break;
              }
            case 'R':
              {
                float    fRotation = g_pMunsell->GetRotation();
                fRotation -= 0.1f;             // in rads
                if ( fRotation < 0.0f )
                    fRotation = (float)(fRotation + TWOPI);
                g_pMunsell->SetRotation( fRotation );
                break;
              }
            case 'e':
              {
                float    fElevation = g_pMunsell->GetElevation();
                fElevation += 0.1f;            // in rads
                if ( fElevation > HALFPI - 0.1f )
                    fElevation = (float)(HALFPI - 0.1f );
                g_pMunsell->SetElevation( fElevation );
                break;
              }
            case 'E':
              {
                float    fElevation = g_pMunsell->GetElevation();
                fElevation -= 0.1f;            // in rads
                if ( fElevation < -HALFPI + 0.1f )
                    fElevation = (float)(-HALFPI + 0.1f);
                g_pMunsell->SetElevation( fElevation );
                break;
              }
            case 'z':
              {
                float   fViewRadius = g_pMunsell->GetViewRadius();
                fViewRadius -= 0.1f;
                if ( fViewRadius < 0.1f )
                    fViewRadius = 0.1f;
                g_pMunsell->SetViewRadius( fViewRadius );
                break;
              }
            case 'Z':
              {
                float   fViewRadius = g_pMunsell->GetViewRadius();
                fViewRadius += 0.1f;
                if ( fViewRadius > 25.0f )
                    fViewRadius = 25.0f;
                g_pMunsell->SetViewRadius( fViewRadius );
                break;
              }
            }
    
            Render();
            break;
        }
        case WM_SETFOCUS:
        {
            HRESULT   hr;
            hr = Render();
            return 0;
        }
        case WM_CLOSE:
        {
            DestroyWindow( hWnd );     // automatically destroys all child
                                       //   windows owned by this parent
            return 0;
        }
        case WM_DESTROY:
        {
            PostQuitMessage( 0 );
            return 0; 
        }
    } // end of switch ( msg )

    return DefWindowProc( hWnd, msg, wParam, lParam);

} // WndProc()


///////////////////////////////////////////////////////////////////////////////
// InitDirect3D()
//
///////////////////////////////////////////////////////////////////////////////

HRESULT InitDirect3D( HWND hWndFocus )
{
    // COM objects are not normal C++ objects and hence are not
    // created using the "new" keyword.  With C++ objects, you
    // can control the object's lifetime directly with the new
    // and delete operators. COM does not enable you to create
    // or destroy objects directly. The reason for this practice
    // is that the same object may be used by more than one application.
    // The lifetime of a COM object is managed via a "reference count"
    // which is the number of times one of its interfaces has been
    // requested.  When the reference count reaches zero, the
    // object destroys itself (and hence is removed from memory).
    // To avoid a memory leak, call IUnknown::Release() to decrement
    // the reference count when you no longer need an interface pointer.
    // All object creation methods (direct and indirect) automatically
    // call IUnknown::AddRef() for you when they return you an
    // interface pointer.  But you will need to make an explicit
    // call to AddRef() should you ever copy a COM interface pointer.

    // The Direct3D object is the first COM object your application
    // creates and the last object your application releases.
    // We use the Direct3DCreate8() function to retrieve a pointer
    // to the Direct3D interface.

    // Note that if the computer does not have Direct3D 8 installed
    // then (for both Release and Debug versions) the code will never
    // even reach this point.  Instead the Windows OS will put up a
    // message box with the contents:
    //         "A required .DLL file, D3D8.DLL, was not found"

    if ( g_pD3D )
        SAFE_RELEASE( g_pD3D);

    g_pD3D = Direct3DCreate8( D3D_SDK_VERSION );

    if ( g_pD3D == NULL )
        return E_FAIL;

    // The IDirect3D8::GetDeviceCaps() method now allows you to query
    // for the capabilities of a device without requiring you to
    // actually create the device, which is a vast simplification for
    // caps querying over previous versions of Direct3D.

    HRESULT    hr;

    do                    // this do while loop just facilitates the
    {                     //    if ( FAILED( hr ) ) break;
                          // statements

        // We need to get the parameters of the current video display mode
        // so we can make our back buffer compatible with it.
        // Rather than enumerating the display adapters on the system
        // [using IDirect3D8::EnumAdapterModes()] we just accept the
        // default display adapter.

        // Since we want to employ Direct3D within a window, we
        // need to call IDirect3D8::CheckDeviceType() to determine
        // whether the device can support rendering to a back buffer
        // of the specified format while in desktop mode.
        // IDirect3D8::GetAdapterDisplayMode() will return the desktop
        // display format.

        D3DDISPLAYMODE   d3ddm;

        hr = g_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm );

        if ( FAILED( hr ) )
            break;

        // CheckDeviceType() can be used to verify a device's ability
        // to perform hardware acceleration, a device's ability to
        // build a swap chain for presentation, or a device's ability
        // to render to the current display format.

        hr = g_pD3D->CheckDeviceType( D3DADAPTER_DEFAULT,
                                      D3DDEVTYPE_HAL,
                                      d3ddm.Format,
                                      d3ddm.Format,
                                      false );
       
        if ( FAILED( hr ) )
            break;

        // D3DPRESENT_PARAMETERS describes information such as the format
        // of a displays surfaces, the type of swapping mechanism, and
        // whether the app is windowed or full screen.

        D3DPRESENT_PARAMETERS   d3dpp;

        ZeroMemory( &d3dpp, sizeof( d3dpp ) );

        // Note that currently d3dpp.BackBufferCount = 0 which
        // is treated as 1.

        d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
                                      // this is the fastest option because it
                                      //   does not guarantee the back buffer
                                      //   will be unaffected by the flip/copy
                        // D3DSWAPEFFECT_COPY;
                                      // only legal when the swap chain has a
                                      //   single back buffer
                        // D3DSWAPEFFECT_COPY_VSYNC;
                                      // by synchronizing with the vertical sync
                                      //   we minimize tearing
                        // D3DSWAPEFFECT_FLIP;
        d3dpp.Windowed = true;
        d3dpp.BackBufferFormat = d3ddm.Format;
                  // If we were going to operate Direct3D in full screen
                  // mode then we could select a back buffer format
                  // such as D3DFMT_X1R5G5B5.
        d3dpp.BackBufferCount = 1;

        // Because we called ZeroMemory(&d3dpp,...) the
        // .BackBufferWidth and .BackBufferHeight members of
        // the D3DPRESENT_PARAMETERS struct are zero.  Microsoft's
        // documentation says that in this situation the dimensions
        // of the focus window are used to size the backbuffer.
        // This means we will have to resize (reset) the Direct3D device
        // whenever the window changes size.  Otherwise Direct3D will
        // use a stretch BLT to fit its output within the new window and
        // this may look bad.

        d3dpp.EnableAutoDepthStencil = true;
                  // "auto" indicates that the depth buffer will be managed
                  // by the system.  When the device is reset, the depth
                  // buffer will be automatically destroyed and recreated
                  // in the new size.
        d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
                  // 16 bit Z buffer (the most commonly supported bit-depth)

        // IDirect3DDevice8::CreateDevice() simplifies device creation
        // by including many tasks (such as back-buffer creation,
        // the equivalent of SetDisplayMode(), specifying what "Present"
        // means, choosing an acceleration level, specifying buffer
        // formats, etc.) that previously required separate calls.

        //          *********  WARNING  *********
        // IDirect3D8::CreateDevice() should not be called during
        // the handling of the WM_CREATE message.

        hr = g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, // use primary monitor
                   // IDirect3D::GetAdapterCount() will return the
                   // number of monitors in the system.
                                   D3DDEVTYPE_HAL,     // create a HAL device 
                                   hWndFocus,
                   // The Direct3D run time hooks this window and
                   // thereby manages to inspect all Windows messages
                   // even before this window's WndProc().
                                   D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                   // You can select software or hardware vertex processing or
                   // a combination of the two. Software vertex processing
                   // gives maximum compatibility with various video cards
                   // while hardware vertex processing gives a considerable
                   // performance boost if your video card supports hardware
                   // transform and lighting (if it doesn't, hardware
                   // and software vertex processing have similar performance).
                                   &d3dpp,
                                   &g_pd3dDevice );

        // If Direct3D8::CreateDevice() returns D3DERR_NOTAVAILABLE
        // then you passed in valid parameters but the device does not
        // support them.  This function automatically creates all your
        // needed back buffers and depth buffers. Clipping is automatically
        // enabled, as is backface culling and lighting.

        // Note that a call to create, release, or reset the device should
        // happen only on the same thread as the window procedure of the
        // focus window.

        if ( FAILED( hr ) )
        {
            MessageBoxFmt( NULL, g_szAppName, MB_ICONERROR,
                           TEXT( "IDirect3D8::CreateDevice() returned the " )
                           TEXT( "HRESULT 0x%x:\n  %s\n" ),
                           hr, DXGetErrorString8( hr ) );
            break;
        }

    } while ( 0 );

    return hr;

} // InitDirect3D()


///////////////////////////////////////////////////////////////////////////////
// ShutdownDirect3D()
//
///////////////////////////////////////////////////////////////////////////////

void ShutdownDirect3D()
{
    SAFE_RELEASE( g_pd3dDevice );
    SAFE_RELEASE( g_pD3D );

    return;

} // ShutdownDirect3D()


///////////////////////////////////////////////////////////////////////////////
// RestoreDeviceObjects()
//
// RestoreDeviceObjects() is used to recreate objects after a Reset() on
// the D3DDevice.  Resources in D3DPOOL_DEFAULT and any other device state
// that persists during rendering should be set here. Render states,
// matrices, textures, etc., that don't change during rendering can be
// set once here to avoid redundant state setting during Render(). 
///////////////////////////////////////////////////////////////////////////////

HRESULT RestoreDeviceObjects()
{
    HRESULT      hr;

    hr = g_pMunsell->CreateGeometry();  // creates vertex and index buffers

    hr = g_pMunsell->SetRenderStates();

    hr = g_pFont->RestoreDeviceObjects();

    return hr;

} // RestoreDeviceObjects()


///////////////////////////////////////////////////////////////////////////////
//  Render()
//
///////////////////////////////////////////////////////////////////////////////

HRESULT Render()
{
    HRESULT    hr;

    // Test the cooperative level to see if it's okay to render.

    if ( FAILED( hr = g_pd3dDevice->TestCooperativeLevel() ) )
    {
        // If the device was lost, do not render until we get it back.

        if ( D3DERR_DEVICELOST == hr )
            return S_OK;

        // Check if the device needs to be resized.

        if ( D3DERR_DEVICENOTRESET == hr )
        {
            if ( FAILED( hr = Resize3DEnvironment() ) )
                              // calls IDirect3DDevice8::Reset()
            {
                return hr;
            }
        }
    }

    hr = g_pMunsell->Render();

    return hr;

} // Render()


///////////////////////////////////////////////////////////////////////////////
//  Resize3DEnvironment()
//
// This function is called in response to the WM_SIZE message.
///////////////////////////////////////////////////////////////////////////////

HRESULT Resize3DEnvironment()
{
    HRESULT      hr;

    // Release all device-dependent objects.  [In Microsoft's SDK
    // sample programs the following code would be placed in an
    // InvalidateDeviceObjects() function.]

    hr = g_pFont->InvalidateDeviceObjects();

    hr = g_pMunsell->ReleaseGeometry();  // releases vertex and index buffers

    // Reset the device.  Calling IDirect3DDevice8::Reset() causes all
    // texture memory surfaces to be lost, managed textures to be
    // flushed from video memory, and all state information to be lost.
    // Before calling the Reset() method for a device, an application
    // should release any explicit render targets, depth stencil surfaces,
    // additional swap chains and D3DPOOL_DEFAULT resources associated
    // with the device.

    D3DDISPLAYMODE   d3ddm;

    hr = g_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm );

    if ( FAILED( hr ) )
        return hr;

    D3DPRESENT_PARAMETERS   d3dpp;

    ZeroMemory( &d3dpp, sizeof( d3dpp ) );
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.Windowed = true;
    d3dpp.BackBufferFormat = d3ddm.Format;
    d3dpp.BackBufferCount = 1;
    d3dpp.EnableAutoDepthStencil = true;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
    // Because we called ZeroMemory(&d3dpp,...) the
    // .BackBufferWidth and .BackBufferHeight members of
    // the D3DPRESENT_PARAMETERS struct are zero.  Microsoft's
    // documentation says that in this situation the dimensions
    // of the focus window are used to size the backbuffer.
    // This means we will have to resize (reset) the Direct3D device
    // whenever the window changes size (otherwise Direct3D
    // will use a stretch BLT to fit its output within the new
    // window and this may look bad).

    if ( FAILED( hr = g_pd3dDevice->Reset( &d3dpp ) ) )
        return hr;

    // Initialize the app's device-dependent objects.

    hr = RestoreDeviceObjects();     // calls cMunsell::CreateGeometry()

    if ( FAILED( hr ) )
        return hr;

    return hr;

} // Resize3DEnvironment()


//////////////////////////////////////////////////////////////////////////////
//  MessageBoxFmt()
//
//  This function works like printf() in that it supports a variable
//  argument list.  But the string that it constructs is displayed
//  via a call to MessageBox().  The message box has a single
//  pushbutton of the specified type uiStyle (e.g., MB_OK, MB_ICONERROR,
//  etc.) which the user must click to dismiss the window.  The only
//  consequence of the hWnd passed parameter is that the message box
//  will be constrained to lie within this parent window.
//
//  Because this function uses a variable argument list must be declared
//  as using the CDECL calling convention.  Furthermore, the program must
//  #include <stdarg.h> .
//////////////////////////////////////////////////////////////////////////////

int CDECL MessageBoxFmt( HWND hWnd,               // parent window
                         const TCHAR * pszTitle,  // title for message
                         UINT uiStyle,            // style of single pushbutton
                         const TCHAR * pszFormat, // format string
                         ... )                    // substitution parameters
{
    static TCHAR  szBuffer[BUFSIZ];
    va_list       pArgList;
    UINT          uiMaxChars = sizeof( szBuffer ) / sizeof( TCHAR );
    UINT          uiNumWritten;

    va_start( pArgList, pszFormat );
    uiNumWritten = _vsntprintf( szBuffer, uiMaxChars, pszFormat, pArgList );
       //  _vsntprintf() will write no more than uiMaxChars characters,
       //  which guarantees that buffer won't be overrun but doesn't
       //  guarantee that the string will be NULL terminated.  So ...
    szBuffer[uiMaxChars-1] = TEXT( '\0' );
       //  Note that some flavors of printf() such as wvsprintf() can't
       //  handle floating point numbers and hence will mis-print a
       //  %f or %lf format speciier.
    va_end( pArgList );

    return MessageBox( hWnd, szBuffer, pszTitle, uiStyle );

} // MessageBoxFmt()


